var xtend = require('xtend');
var jwt = require('jsonwebtoken');
var UnauthorizedError = require('./UnauthorizedError');
var debug = require('debug')('socketio-jwt-auth');

function noQsMethod(options) {
  var defaults = { required: true };
  options = xtend(defaults, options);

  return function (socket, next) {
    middleware(socket, options);
    var server = this.server || socket.server;

    if (!server.$emit) {
      //then is socket.io 1.0
      var Namespace = Object.getPrototypeOf(server.sockets).constructor;
      if (!~Namespace.events.indexOf('authenticated')) {
        Namespace.events.push('authenticated');
      }
    }

    if(options.required){
      var auth_timeout = setTimeout(function () {
        socket.disconnect('unauthorized');
      }, options.timeout || 5000);
    }

    socket.on('signup', signup.bind({
      socket: socket,
      options: options
    }));

    socket.on('signin', signin.bind({
      socket: data,
      options: options
    }));

    socket.on('signout', signout.bind({
      socket: data,
      options: options
    }));


    socket.on('authenticate', function (data) {
      if(options.required){
        clearTimeout(auth_timeout);
      }
      // error handler
      var onError = function(err, code) {
          if (err) {
            code = code || 'unknown';
            var error = new UnauthorizedError(code, {
              message: (Object.prototype.toString.call(err) === '[object Object]' && err.message) ? err.message : err
            });
            var callback_timeout;
            // If callback explicitely set to false, start timeout to disconnect socket
            if (options.callback === false || typeof options.callback === "number") {
              if (typeof options.callback === "number") {
                if (options.callback < 0) {
                  // If callback is negative(invalid value), make it positive
                  options.callback = Math.abs(options.callback);
                }
              }
              callback_timeout = setTimeout(function () {
                socket.disconnect('unauthorized');
              }, (options.callback === false ? 0 : options.callback));
            }
            socket.emit('unauthorized', error, function() {
              if (typeof options.callback === "number") {
                clearTimeout(callback_timeout);
              }
              socket.disconnect('unauthorized');
            });
            return; // stop logic, socket will be close on next tick
          }
      };

      if(!data || typeof data.token !== "string") {
        return onError({message: 'invalid token datatype'}, 'invalid_token');
      }

      var onJwtVerificationReady = function(err, decoded) {

        if (err) {
          return onError(err, 'invalid_token');
        }

        // success handler
        var onSuccess = function() {
          socket[options.decodedPropertyName] = decoded;
          socket.emit('authenticated');
          if (server.$emit) {
            server.$emit('authenticated', socket);
          } else {
            //try getting the current namespace otherwise fallback to all sockets.
            var namespace = (server.nsps && socket.nsp &&
                             server.nsps[socket.nsp.name]) ||
                            server.sockets;

            // explicit namespace
            namespace.emit('authenticated', socket);
          }
        };

        if(options.additional_auth && typeof options.additional_auth === 'function') {
          options.additional_auth(decoded, onSuccess, onError);
        } else {
          onSuccess();
        }
      };

      var onSecretReady = function(err, secret) {
        if (err || !secret) {
          return onError(err, 'invalid_secret');
        }

        jwt.verify(data.token, secret, options, onJwtVerificationReady);
      };

      getSecret(socket.request, options.secret, data.token, onSecretReady);
    });

    return next();
  };
}

function authorize(options, onConnection) {
  debug('< authorize', options)
  options = xtend({ decodedPropertyName: 'decoded_token' }, options);

  if (!options.handshake) {
    return noQsMethod(options);
  }

  var defaults = {
    success: function(data, accept){
      if (data.request) {
        accept();
      } else {
        accept(null, true);
      }
    },
    // will authorize now in any case
    fail: function(error, data, accept){
      if (data.request) {
        accept();
      } else {
        accept(null, true);
      }
    }
  };

  var auth = xtend(defaults, options);

  return function(data, accept) {
    middleware(data, options);

    data.on('signup', signup.bind({
      socket: data,
      options: options
    }));

    data.on('signin', signin.bind({
      socket: data,
      options: options
    }));

    data.on('signout', signout.bind({
      socket: data,
      options: options
    }));

    var token, error;
    var req = data.request || data;
    var authorization_header = (req.headers || {}).authorization;

    if (authorization_header) {
      var parts = authorization_header.split(' ');
      if (parts.length == 2) {
        var scheme = parts[0],
          credentials = parts[1];

        if (scheme.toLowerCase() === 'bearer') {
          token = credentials;
        }
      } else {
        error = new UnauthorizedError('credentials_bad_format', {
          message: 'Format is Authorization: Bearer [token]'
        });
        return auth.fail(error, data, accept);
      }
    }

    //get the token from query string
    if (req._query && req._query.token) {
      token = req._query.token;
    }
    else if (req.query && req.query.token) {
      token = req.query.token;
    }

    if (!token) {
      error = new UnauthorizedError('credentials_required', {
        message: 'No Authorization header was found'
      });
      return auth.fail(error, data, accept);
    }

    var onJwtVerificationReady = function(err, decoded) {

      if (err) {
        error = new UnauthorizedError(err.code || 'invalid_token', err);
        return auth.fail(error, data, accept);
      }

      data[options.decodedPropertyName] = decoded;

      return auth.success(data, accept);
    };

    var onSecretReady = function(err, secret) {
      if (err) {
        error = new UnauthorizedError(err.code || 'invalid_secret', err);
        return auth.fail(error, data, accept);
      }

      jwt.verify(token, secret, options, onJwtVerificationReady);
    };

    getSecret(req, options.secret, token, onSecretReady);
  };
}

function getSecret(request, secret, token, callback) {
  if (typeof secret === 'function') {
    if (!token) {
      return callback({ code: 'invalid_token', message: 'jwt must be provided' });
    }

    var parts = token.split('.');

    if (parts.length < 3) {
      return callback({ code: 'invalid_token', message: 'jwt malformed' });
    }

    if (parts[2].trim() === '') {
      return callback({ code: 'invalid_token', message: 'jwt signature is required' });
    }

    var decodedToken = jwt.decode(token);

    if (!decodedToken) {
      return callback({ code: 'invalid_token', message: 'jwt malformed' });
    }

    secret(request, decodedToken, callback);
  } else {
    callback(null, secret);
  }
};

function signup(data) {
  var socket = this.socket;
  var options = this.options;

  debug('< signup request', data)

  // error handler
  var onError = function(err, code) {
    if (err) {
      code = code || 'unknown';
      var error = new UnauthorizedError(code, {
        message: (Object.prototype.toString.call(err) === '[object Object]' && err.message) ? err.message : err
      });
      socket.emit('denied', error);
      return;
    }
  };

  if (!options.validateUsername || !options.createUser) {
    return onError({message: 'invalid signup configuration'}, 'missing user callbacks');
  }

  if (!data) {
    return onError({message: 'invalid signup request'}, 'empty');
  }

  if (!data.username || !data.password) {
    return onError({message: 'invalid signup request'}, 'data malformed');
  }

  // validateUsername must be synchronous (use externally await/async if necessary)
  if (!options.validateUsername(data.username)) {
    return onError({message: 'invalid signup request'}, 'username already in use');
  }

  // createUser must be synchrounous (use externally await/async if necessary)
  var user = options.createUser(data);
  if (user) {
    if (options.secret === 'function') {
      // API modified decodedToken can be an user
      secret = options.secret(socket.request, user, function(err, computed) {
        socket.emit('token', {
          token: jwt.sign(user, computed, { expiresIn: options.tokenTimeout || '5h' })
        });
      });
    } else {
      socket.emit('token', {
        token: jwt.sign(user, options.secret, { expiresIn: options.tokenTimeout || '5h' })
      });
    }
    debug('< new user', user)
  }
};


function signin(data) {
  var socket = this.socket;
  var options = this.options;

  debug('< signin request', data)

  // error handler
  var onError = function(err, code) {
    if (err) {
      code = code || 'unknown';
      var error = new UnauthorizedError(code, {
        message: (Object.prototype.toString.call(err) === '[object Object]' && err.message) ? err.message : err
      });
      socket.emit('denied', error);
      return;
    }
  };

  if (!options.validateUsername) {
    return onError({message: 'invalid signup configuration'}, 'missing user callbacks');
  }

  if (!data) {
    return onError({message: 'invalid signin request'}, 'empty');
  }

  if (!data.username || !data.password) {
    return onError({message: 'invalid signin request'}, 'data malformed');
  }

  // validateUser must be synchronous (use externally await/async if necessary)
  if (!options.validateUser(data)) {
    return onError({message: 'singin unauthorized'}, 'user info does not match any entry');
  }

  socket.emit('logged', data.username);
  socket.logged = true;

  debug('< user logged',  data.username)

};

function signout(data) {
  var socket = this.socket;
  var options = this.options;

  debug('< signout request', data)

  // error handler
  var onError = function(err, code) {
    if (err) {
      code = code || 'unknown';
      var error = new UnauthorizedError(code, {
        message: (Object.prototype.toString.call(err) === '[object Object]' && err.message) ? err.message : err
      });
      socket.emit('denied', error);
      return;
    }
  };

  if (!data) {
    return onError({message: 'invalid signout request'}, 'empty');
  }

  if (!data.username) {
    return onError({message: 'invalid signout request'}, 'data malformed');
  }

  socket.emit('unlogged', data.username);
  socket.logged = false;

  debug('< user unlogged',  data.username)

};

function middleware(socket, options) {
  var onevent = socket.onevent;
  socket.onevent = function (packet) {
      var args = packet.data || [];
      if (packet.data.length == 0) packet.data.push('_internal');
      if (packet.data.length == 1) packet.data.push({});
      packet.data[1]._auth = socket[options.decodedPropertyName] || socket.logged ? true : false;
      onevent.call (this, packet);    // original call
      /*packet.data = ["*"].concat(args);
      onevent.call(this, packet);      // additional call to catch-all*/
  };
};

exports.authorize = authorize;
